@@ -1,4 +1,4 @@
-# Copyright 2021-2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -23,7 +23,7 @@ import traceback
 from collections.abc import Iterator
 from pathlib import Path
 from typing import Any, cast
-from unittest import TestCase, mock
+from unittest import mock
 from unittest.mock import MagicMock
 
 import psutil
@@ -34,13 +34,14 @@ except ImportError:
     import pydantic as pydantic  # type: ignore
 
 from debusine.artifacts import WorkRequestDebugLogs
-from debusine.artifacts.models import ArtifactCategory
+from debusine.artifacts.models import ArtifactCategory, EmptyArtifactData
 from debusine.client.debusine import Debusine
 from debusine.client.models import ArtifactResponse
 from debusine.tasks import (
     BaseExternalTask,
     BaseTask,
-    BaseTaskWithExecutor,
+    DefaultDynamicData,
+    ExtraRepositoryMixin,
     RunCommandTask,
     Sbuild,
     TaskConfigError,
@@ -56,20 +57,28 @@ from debusine.tasks.models import (
     BaseDynamicTaskDataWithExecutor,
     BaseTaskData,
     BaseTaskDataWithExecutor,
+    BaseTaskDataWithExtraRepositories,
+    ExtraRepository,
     TaskTypes,
     WorkerType,
 )
+from debusine.tasks.server import ArtifactInfo
 from debusine.tasks.tests.helper_mixin import (
     ExternalTaskHelperMixin,
     FakeTaskDatabase,
-    TestTaskMixin,
+    TestBaseExternalTask,
+    TestBaseTask,
+    TestBaseTaskWithExecutor,
+)
+from debusine.test import TestCase
+from debusine.test.test_utils import (
+    create_system_tarball_data,
+    preserve_task_registry,
 )
-from debusine.test import TestHelpersMixin
-from debusine.test.utils import preserve_task_registry
 
 
-class TestBaseTask(
-    TestTaskMixin, BaseExternalTask[BaseTaskData, BaseDynamicTaskData]
+class TestBaseTask1(
+    TestBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
 ):
     """Sample class to test BaseExternalTask class."""
 
@@ -80,9 +89,8 @@ class TestBaseTask(
         raise NotImplementedError()
 
 
-class TestBaseTaskWithExecutor(
-    TestTaskMixin,
-    BaseTaskWithExecutor[
+class TestBaseTaskWithExecutor1(
+    TestBaseTaskWithExecutor[
         BaseTaskDataWithExecutor, BaseDynamicTaskDataWithExecutor
     ],
 ):
@@ -106,7 +114,7 @@ class TestBaseTask2Data(BaseTaskData):
 
 
 class TestBaseTask2(
-    TestTaskMixin, BaseExternalTask[TestBaseTask2Data, BaseDynamicTaskData]
+    TestBaseExternalTask[TestBaseTask2Data, BaseDynamicTaskData],
 ):
     """Test BaseExternalTask class with data validation."""
 
@@ -119,9 +127,7 @@ class TestBaseTask2(
         raise NotImplementedError()
 
 
-class TestBaseTaskDataWithExecutorAndArchitecture(
-    TestTaskMixin, BaseTaskDataWithExecutor
-):
+class TestBaseTaskDataWithExecutorAndArchitecture(BaseTaskDataWithExecutor):
     """BaseTaskDataWithExecutor with a host architecture."""
 
     __test__ = False
@@ -130,8 +136,7 @@ class TestBaseTaskDataWithExecutorAndArchitecture(
 
 
 class TestBaseTaskWithExecutorAndArchitecture(
-    TestTaskMixin,
-    BaseTaskWithExecutor[
+    TestBaseTaskWithExecutor[
         TestBaseTaskDataWithExecutorAndArchitecture,
         BaseDynamicTaskDataWithExecutor,
     ],
@@ -147,14 +152,46 @@ class TestBaseTaskWithExecutorAndArchitecture(
         raise NotImplementedError()
 
 
-class BaseTaskTests(TestHelpersMixin, TestCase):
+class TestExtraRepositoryMixin(
+    TestBaseTask[BaseTaskDataWithExtraRepositories, BaseDynamicTaskData],
+    ExtraRepositoryMixin[
+        BaseTaskDataWithExtraRepositories, BaseDynamicTaskData
+    ],
+):
+    """Test ExtraRepositoryMixin."""
+
+    def run(self, execute_directory: Path) -> bool:
+        """Unused abstract method from BaseExternalTask."""
+        raise NotImplementedError()
+
+
+class TestDefaultDynamicData(
+    DefaultDynamicData[BaseTaskData],
+    TestBaseTask[BaseTaskData, BaseDynamicTaskData],
+):
+    """Test DefaultDynamicData."""
+
+    TASK_TYPE = TaskTypes.WORKER
+
+    def _execute(self) -> bool:
+        raise NotImplementedError()
+
+    def _upload_work_request_debug_logs(self) -> None:
+        raise NotImplementedError()
+
+    def run(self, execute_directory: Path) -> bool:
+        """Unused abstract method from BaseExternalTask."""
+        raise NotImplementedError()
+
+
+class BaseTaskTests(TestCase):
     """Unit tests for BaseTask class."""
 
     worker_metadata = {"system:worker_type": WorkerType.EXTERNAL}
 
     def setUp(self) -> None:
         """Create the shared attributes."""
-        self.task = TestBaseTask({})
+        self.task = TestBaseTask1({})
         self.task2 = TestBaseTask2({"foo": "bar"})
 
     def tearDown(self) -> None:
@@ -169,7 +206,7 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
 
     def test_name_is_set(self) -> None:
         """task.name is built from class name."""
-        self.assertEqual(self.task.name, "testbasetask")
+        self.assertEqual(self.task.name, "testbasetask1")
 
     def test_logger_is_configured(self) -> None:
         """task.logger is available."""
@@ -254,6 +291,18 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
             TestWait.prefix_with_task_name("foo"), "wait:testwait:foo"
         )
 
+    def test_get_subject(self) -> None:
+        self.task.dynamic_data = BaseDynamicTaskDataWithExecutor(
+            subject="hello"
+        )
+
+        self.assertEqual(self.task.get_subject(), "hello")
+
+    def test_get_subject_dynamic_task_data_is_none(self) -> None:
+        self.task.dynamic_data = None
+
+        self.assertIsNone(self.task.get_subject())
+
     def test_configure_sets_data_attribute(self) -> None:
         """task.data is set with configure."""
         task_data = {
@@ -297,7 +346,7 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
     def test_analyze_worker_without_task_version(self) -> None:
         """analyze_worker() reports an unknown task version."""
         metadata = self.task.analyze_worker()
-        self.assertIsNone(metadata["testbasetask:version"])
+        self.assertIsNone(metadata["testbasetask1:version"])
 
     def test_analyze_worker_with_task_version(self) -> None:
         """analyze_worker() reports a task version."""
@@ -310,7 +359,10 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
             BaseTask,
             "_sub_tasks",
             new={
-                TaskTypes.WORKER: {"task": TestBaseTask, "task2": TestBaseTask2}
+                TaskTypes.WORKER: {
+                    "task": TestBaseTask1,
+                    "task2": TestBaseTask2,
+                }
             },
         )
         patcher.start()
@@ -380,15 +432,13 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
         with self.assertRaisesRegex(AssertionError, msg):
 
             class Somesubclassname(
-                TestTaskMixin,
-                BaseExternalTask[BaseTaskData, BaseDynamicTaskData],
+                TestBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
             ):
                 def run(self, execute_directory: Path) -> bool:  # noqa: U100
                     raise NotImplementedError()
 
             class SomeSubclassName(
-                TestTaskMixin,
-                BaseExternalTask[BaseTaskData, BaseDynamicTaskData],
+                TestBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
             ):
                 def run(self, execute_directory: Path) -> bool:  # noqa: U100
                     raise NotImplementedError()
@@ -400,7 +450,7 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
         with self.assertRaisesRegex(AssertionError, msg):
 
             class Somesubclassname(
-                TestTaskMixin, BaseTask[BaseTaskData, BaseDynamicTaskData]
+                TestBaseTask[BaseTaskData, BaseDynamicTaskData]
             ):
                 TASK_TYPE = TaskTypes.SERVER
 
@@ -411,8 +461,7 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
                     raise NotImplementedError()
 
             class SomeSubclassName(
-                TestTaskMixin,
-                BaseExternalTask[BaseTaskData, BaseDynamicTaskData],
+                TestBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
             ):
                 def run(self, execute_directory: Path) -> bool:  # noqa: U100
                     raise NotImplementedError()
@@ -421,14 +470,13 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
         with self.assertRaisesRegex(AssertionError, msg):
 
             class SomeSubclassName1(
-                TestTaskMixin,
-                BaseExternalTask[BaseTaskData, BaseDynamicTaskData],
+                TestBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
             ):
                 def run(self, execute_directory: Path) -> bool:  # noqa: U100
                     raise NotImplementedError()
 
             class Somesubclassname1(
-                TestTaskMixin, BaseTask[BaseTaskData, BaseDynamicTaskData]
+                TestBaseTask[BaseTaskData, BaseDynamicTaskData]
             ):
                 TASK_TYPE = TaskTypes.SERVER
 
@@ -438,7 +486,7 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
                 def _upload_work_request_debug_logs(self) -> None:
                     raise NotImplementedError()
 
-    def test_init_subclass_no_register_abstract_class(self):
+    def test_init_subclass_no_register_abstract_class(self) -> None:
         """BaseTask.__init_subclass__ does not register abstract classes."""
 
         class Abstract(BaseTask[BaseTaskData, BaseDynamicTaskData], abc.ABC):
@@ -625,13 +673,13 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
     def test_task_names_worker(self) -> None:
         """BaseTask.task_names() returns all available worker tasks."""
         task_names = self.task.task_names(TaskTypes.WORKER)
-        self.assertTrue(len(task_names) > 0)
+        self.assertGreater(len(task_names), 0)
         self.assertIn("noop", task_names)
 
     def test_worker_task_names(self) -> None:
         """BaseTask.worker_task_names() returns all available worker tasks."""
         task_names = self.task.worker_task_names()
-        self.assertTrue(len(task_names) > 0)
+        self.assertGreater(len(task_names), 0)
         self.assertIn("noop", task_names)
 
     def test_is_worker_task(self) -> None:
@@ -642,9 +690,67 @@ class BaseTaskTests(TestHelpersMixin, TestCase):
         """BaseTask.is_worker_task(): nonexistent isn't a WORKER task."""
         self.assertFalse(self.task.is_worker_task("nonexistent"))
 
+    def test_ensure_artifact_categories_artifact(self) -> None:
+        BaseTask.ensure_artifact_categories(
+            configuration_key="source_artifact",
+            category=ArtifactCategory.TEST,
+            expected=[ArtifactCategory.TEST, ArtifactCategory.SOURCE_PACKAGE],
+        )
 
-class TestBaseExternalTask(
-    TestTaskMixin, BaseExternalTask[BaseTaskData, BaseDynamicTaskData]
+    def test_ensure_artifact_categories_raise(self) -> None:
+        with self.assertRaisesRegex(
+            TaskConfigError,
+            r"^source_artifact: unexpected artifact category: 'debusine:test'. "
+            r"Valid categories: "
+            r"\['debian:binary-package', 'debian:binary-packages'\]$",
+        ):
+            BaseTask.ensure_artifact_categories(
+                configuration_key="source_artifact",
+                category=ArtifactCategory.TEST,
+                expected=[
+                    ArtifactCategory.BINARY_PACKAGE,
+                    ArtifactCategory.BINARY_PACKAGES,
+                ],
+            )
+
+    def test_instantiate_with_new_data(self) -> None:
+        task = self.task2.instantiate_with_new_data({"foo": "baz"})
+        self.assertEqual(self.task2.data.foo, "bar")
+        self.assertEqual(task.data.foo, "baz")
+
+        self.assertIsNone(task.work_request_id)
+        self.assertIsNone(task.workspace_name)
+        self.assertIsNone(task.worker_host_architecture)
+
+    def test_instantiate_with_new_data_copy_attrs(self) -> None:
+        """instantiate_with_new_data copies extra attributes."""
+        self.task2.work_request_id = 42
+        self.task2.workspace_name = "test_workspace"
+        self.task2.worker_host_architecture = "arm64"
+
+        task = self.task2.instantiate_with_new_data({"foo": "bar"})
+
+        self.assertEqual(task.work_request_id, 42)
+        self.assertEqual(task.workspace_name, "test_workspace")
+        self.assertEqual(task.worker_host_architecture, "arm64")
+
+    def test_instantiate_with_new_data_ignores_old_data(self) -> None:
+        """instantiate_with_new_data ignores existing task data."""
+        self.task2.data.bar = 1
+        task = self.task2.instantiate_with_new_data({"foo": "bar"})
+        self.assertEqual(task.data.foo, "bar")
+        self.assertIsNone(task.data.bar)
+
+    def test_compute_dynamic_data(self) -> None:
+        task_db = FakeTaskDatabase()
+
+        self.assertEqual(
+            self.task.compute_dynamic_data(task_db), BaseDynamicTaskData()
+        )
+
+
+class TestBaseExternalTask1(
+    TestBaseExternalTask[BaseTaskData, BaseDynamicTaskData]
 ):
     """Sample class to test BaseExternalTask class."""
 
@@ -656,7 +762,7 @@ class TestBaseExternalTask(
 
 
 class TestBaseExternalTask2(
-    TestTaskMixin, BaseExternalTask[TestBaseTask2Data, BaseDynamicTaskData]
+    TestBaseExternalTask[TestBaseTask2Data, BaseDynamicTaskData]
 ):
     """Test BaseExternalTask class with data validation."""
 
@@ -669,12 +775,12 @@ class TestBaseExternalTask2(
         raise NotImplementedError()
 
 
-class BaseExternalTaskTests(TestHelpersMixin, TestCase):
+class BaseExternalTaskTests(TestCase):
     """Unit tests for :class:`BaseExternalTask`."""
 
     def setUp(self) -> None:
         """Create the shared attributes."""
-        self.task = TestBaseExternalTask({})
+        self.task = TestBaseExternalTask1({})
         self.task2 = TestBaseExternalTask2({"foo": "bar"})
         self.worker_metadata = {"system:worker_type": WorkerType.EXTERNAL}
 
@@ -693,7 +799,7 @@ class BaseExternalTaskTests(TestHelpersMixin, TestCase):
         """Ensure can_run_on returns False if versions differ."""
         self.assertIsNone(self.task.TASK_VERSION)
         metadata = {**self.worker_metadata, **self.task.analyze_worker()}
-        metadata["testbaseexternaltask:version"] = 1
+        metadata["testbaseexternaltask1:version"] = 1
         self.assertFalse(self.task.can_run_on(metadata))
 
     def test_can_run_on_type_mismatch(self) -> None:
@@ -960,8 +1066,12 @@ class BaseExternalTaskTests(TestHelpersMixin, TestCase):
         traceback_str = "".join(traceback.format_exception(exc))
         self.assertIn(traceback_str, log)
 
-    def test_execute_logs_progress(self) -> None:
-        """_execute() logs each of its main steps."""
+    def test_execute_logs_progress_success(self) -> None:
+        """
+        _execute() logs each of its main steps: success.
+
+        For a success, there is no stages.log file.
+        """
         self.task.work_request_id = 1
 
         with (
@@ -988,6 +1098,67 @@ class BaseExternalTaskTests(TestHelpersMixin, TestCase):
                 "INFO:debusine.tasks:Work request 1: Cleaning up",
             ],
         )
+        assert self.task._debug_log_files_directory
+        stages_log = (
+            Path(self.task._debug_log_files_directory.name) / "stages.log"
+        )
+        self.assertFalse(stages_log.exists())
+
+    def test_execute_logs_progress_failure(self) -> None:
+        """
+        _execute() logs each of its main steps: failure.
+
+        For a failure, there is an additional stages.log file.
+        """
+        self.task.work_request_id = 1
+
+        with (
+            mock.patch.object(self.task, "fetch_input", autospec=True),
+            mock.patch.object(
+                self.task, "configure_for_execution", autospec=True
+            ),
+            mock.patch.object(
+                self.task, "run", autospec=True, return_value=False
+            ),
+            mock.patch.object(self.task, "upload_artifacts", autospec=True),
+            mock.patch.object(self.task, "cleanup", autospec=True),
+            self.assertLogs(self.task.logger) as log,
+        ):
+            self.assertFalse(self.task._execute())
+
+        self.assertEqual(
+            log.output,
+            [
+                "INFO:debusine.tasks:Work request 1: Fetching input",
+                "INFO:debusine.tasks:Work request 1: Configuring for execution",
+                "INFO:debusine.tasks:Work request 1: Preparing to run",
+                "INFO:debusine.tasks:Work request 1: Running",
+                "INFO:debusine.tasks:Work request 1: Checking output",
+                "INFO:debusine.tasks:Work request 1: Uploading artifacts",
+                "INFO:debusine.tasks:Work request 1: Cleaning up",
+            ],
+        )
+        assert self.task._debug_log_files_directory
+        stages_log = (
+            Path(self.task._debug_log_files_directory.name) / "stages.log"
+        )
+        stages_log_lines = []
+        for line in stages_log.read_text().splitlines():
+            dt, message = line.split(maxsplit=1)
+            datetime.datetime.fromisoformat(dt)
+            stages_log_lines.append(message)
+        self.assertEqual(
+            stages_log_lines,
+            [
+                "Fetching input",
+                "Configuring for execution",
+                "Preparing to run",
+                "Running",
+                "Checking output",
+                "Uploading artifacts",
+                "Cleaning up",
+            ],
+        )
 
     def setup_upload_work_request_debug_logs(
         self, source_artifacts_ids: list[int] | None = None
@@ -1071,7 +1242,7 @@ class TestsTaskConfigError(TestCase):
     """Tests for TaskConfigError."""
 
     def test_without_original_exception(self) -> None:
-        """Test TaskConfigError (init and str) without original exception."""
+        """No original exception: pass through message."""
         msg = "This is an error message"
 
         error = TaskConfigError(msg)
@@ -1080,7 +1251,7 @@ class TestsTaskConfigError(TestCase):
         self.assertIsNone(error.original_exception)
 
     def test_str_with_original_exception(self) -> None:
-        """Test TaskConfigError (init and str) with original exception."""
+        """Message and original exception: combine both."""
         msg = "This is an error message"
         original_exc = Exception("something")
 
@@ -1091,9 +1262,28 @@ class TestsTaskConfigError(TestCase):
         )
         self.assertIs(error.original_exception, original_exc)
 
+    def test_None_with_original_exception(self) -> None:
+        """Original exception and no message: stringify exception."""
+        original_exc = Exception("something")
+
+        error = TaskConfigError(None, original_exception=original_exc)
+
+        self.assertEqual(str(error), str(original_exc))
+        self.assertIs(error.original_exception, original_exc)
+
+    def test_add_parent_message(self) -> None:
+        """Test add_parent_message."""
+        for exc, expected in (
+            (TaskConfigError(None), "test"),
+            (TaskConfigError("msg"), "test: msg"),
+        ):
+            exc.add_parent_message("test")
+            self.assertEqual(exc.args[0], expected)
+
 
 class RunCommandTaskForTesting(
-    TestTaskMixin, RunCommandTask[BaseTaskData, BaseDynamicTaskData]
+    TestBaseTask[BaseTaskData, BaseDynamicTaskData],
+    RunCommandTask[BaseTaskData, BaseDynamicTaskData],
 ):
     """Used in TaskTests."""
 
@@ -1229,9 +1419,7 @@ class RunCommandTaskForTesting(
 
 
 class RunCommandTaskTests(
-    TestHelpersMixin,
-    ExternalTaskHelperMixin[RunCommandTaskForTesting],
-    TestCase,
+    ExternalTaskHelperMixin[RunCommandTaskForTesting], TestCase
 ):
     """Tests for RunCommandTask methods in isolation."""
 
@@ -1315,7 +1503,7 @@ class RunCommandTaskTests(
         return mocked
 
     def run_signal_logger(
-        self, task: RunCommandTaskForTesting, options
+        self, task: RunCommandTaskForTesting, options: list[str]
     ) -> int | None:
         """Run signal logger in task with options."""
         # Block SIGUSR1 signal in case that it is delivered before
@@ -1357,7 +1545,7 @@ class RunCommandTaskTests(
         )
 
     @staticmethod
-    def pid_exist_not_zombie(pid) -> bool:
+    def pid_exist_not_zombie(pid: int) -> bool:
         """Return True if the pid exist and is not a zombie process."""
         try:
             status = psutil.Process(pid).status()
@@ -1782,12 +1970,17 @@ class RunCommandTaskTests(
         self.assertEqual(log_file.read_text(), expected_log)
 
 
-class BaseTaskWithExecutorTests(TestHelpersMixin, TestCase):
+class BaseTaskWithExecutorTests(TestCase):
     """Unit tests for BaseTaskWithExecutor class."""
 
     def setUp(self) -> None:
         """Create the shared attributes."""
-        self.task = TestBaseTaskWithExecutor({})
+        self.task = TestBaseTaskWithExecutor1({})
+
+    def tearDown(self) -> None:
+        """Delete directory to avoid ResourceWarning."""
+        if self.task._debug_log_files_directory:
+            self.task._debug_log_files_directory.cleanup()
 
     def patch_executor(
         self, backend: BackendType, image_category: ExecutorImageCategory
@@ -1826,16 +2019,21 @@ class BaseTaskWithExecutorTests(TestHelpersMixin, TestCase):
 
     def test_backend_return_backend(self) -> None:
         """BaseTask.backend return the backend."""
-        backend = BackendType.SCHROOT
+        backend = BackendType.INCUS_LXC
         self.task.data.backend = backend
 
         self.assertEqual(self.task.backend, backend)
 
     def test_get_environment_int(self) -> None:
         """`get_environment` passes through integer lookups."""
-        task_db = FakeTaskDatabase(single_lookups={(1, None): 1})
+        artifact_info = ArtifactInfo(
+            id=1,
+            category=ArtifactCategory.TEST,
+            data=EmptyArtifactData(),
+        )
+        task_db = FakeTaskDatabase(single_lookups={(1, None): artifact_info})
 
-        self.assertEqual(self.task.get_environment(task_db, 1), 1)
+        self.assertEqual(self.task.get_environment(task_db, 1), artifact_info)
 
     def test_get_environment_with_architecture_and_backend(self) -> None:
         """`get_environment` fills in architecture/backend."""
@@ -1848,18 +2046,23 @@ class BaseTaskWithExecutorTests(TestHelpersMixin, TestCase):
         )
         assert task.data.environment
         self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.TARBALL)
+        artifact_info = ArtifactInfo(
+            id=2,
+            category=ArtifactCategory.SYSTEM_TARBALL,
+            data=create_system_tarball_data(),
+        )
         task_db = FakeTaskDatabase(
             single_lookups={
                 (
                     "debian/match:codename=bookworm:architecture=amd64:"
                     "format=tarball:backend=unshare",
                     None,
-                ): 2
+                ): artifact_info
             }
         )
 
         self.assertEqual(
-            task.get_environment(task_db, task.data.environment), 2
+            task.get_environment(task_db, task.data.environment), artifact_info
         )
 
     def test_get_environment_with_architecture_from_worker(self) -> None:
@@ -1873,18 +2076,23 @@ class BaseTaskWithExecutorTests(TestHelpersMixin, TestCase):
         assert task.data.environment
         task.worker_host_architecture = "amd64"
         self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.IMAGE)
+        artifact_info = ArtifactInfo(
+            id=3,
+            category=ArtifactCategory.SYSTEM_TARBALL,
+            data=create_system_tarball_data(),
+        )
         task_db = FakeTaskDatabase(
             single_lookups={
                 (
                     "debian/match:codename=bookworm:architecture=amd64:"
                     "format=image:backend=unshare",
                     None,
-                ): 3
+                ): artifact_info
             }
         )
 
         self.assertEqual(
-            task.get_environment(task_db, task.data.environment), 3
+            task.get_environment(task_db, task.data.environment), artifact_info
         )
 
     def test_get_environment_with_architecture_missing(self) -> None:
@@ -1897,18 +2105,23 @@ class BaseTaskWithExecutorTests(TestHelpersMixin, TestCase):
         )
         assert task.data.environment
         self.patch_executor(BackendType.UNSHARE, ExecutorImageCategory.IMAGE)
+        artifact_info = ArtifactInfo(
+            id=3,
+            category=ArtifactCategory.SYSTEM_TARBALL,
+            data=create_system_tarball_data(),
+        )
         task_db = FakeTaskDatabase(
             single_lookups={
                 (
                     "debian/match:codename=bookworm:format=image:"
                     "backend=unshare",
                     None,
-                ): 3
+                ): artifact_info
             }
         )
 
         self.assertEqual(
-            task.get_environment(task_db, task.data.environment), 3
+            task.get_environment(task_db, task.data.environment), artifact_info
         )
 
     def test_prepare_executor(self) -> None:
@@ -2018,11 +2231,251 @@ class BaseTaskWithExecutorTests(TestHelpersMixin, TestCase):
 
         self.task.executor_instance.stop.assert_called_once()
 
+    def test_run_executor_command(self) -> None:
+        """run_executor_command() runs the command."""
+        self.task.executor_instance = MagicMock(spec=InstanceInterface)
+        self.task.executor_instance.run.return_value.returncode = 0
+        self.task.executor_instance.is_started.return_value = True
+
+        self.task.run_executor_command(["true"], "prepare.log", check=True)
+
+        self.task.executor_instance.run.assert_called_once_with(
+            ["true"],
+            run_as_root=False,
+            stdout=mock.ANY,
+            stderr=mock.ANY,
+            check=True,
+        )
+
+        expected_log = "Executing: true\nExecution completed (exit code 0)\n"
+
+        assert self.task._debug_log_files_directory
+        log_file = (
+            Path(self.task._debug_log_files_directory.name) / "prepare.log"
+        )
+
+        self.assertEqual(log_file.read_text(), expected_log)
+
 
 class TestTestTaskMixin(TestCase):
     """Test TestTaskMixin."""
 
     def test_label(self) -> None:
         """Test get_label."""
-        task = TestTaskMixin()
+        task = TestBaseTask1({})
         self.assertEqual(task.get_label(), "test")
+
+
+class DefaultDynamicDataTest(TestCase):
+    """Test DefaultDynamicData."""
+
+    def test_dynamic_data(self) -> None:
+        """Test build_dynamic_data."""
+        task = TestDefaultDynamicData({})
+        self.assertEqual(
+            task.compute_dynamic_data(FakeTaskDatabase()), BaseDynamicTaskData()
+        )
+
+
+class ExtraRepositoryMixinTests(TestCase):
+    """Unit tests for BaseTaskWithExecutor class."""
+
+    def setUp(self) -> None:
+        """Create the shared attributes."""
+        super().setUp()
+        self.task = TestExtraRepositoryMixin({})
+        self.directory = self.create_temporary_directory()
+
+    def configure_extra_repositories(
+        self, *extra_repositories: dict[str, Any]
+    ) -> None:
+        """Parse a list of dictionaries into ExtraRepository objects."""
+        self.task.data.extra_repositories = []
+        for repo in extra_repositories:
+            self.task.data.extra_repositories.append(
+                ExtraRepository.parse_obj(repo)
+            )
+
+    def test_supports_deb822_sources(self) -> None:
+        """Test supports_deb822_sources with 4 sample distributions."""
+        self.assertFalse(self.task.supports_deb822_sources("jessie"))
+        self.assertFalse(self.task.supports_deb822_sources("trusty"))
+        self.assertTrue(self.task.supports_deb822_sources("stretch"))
+        self.assertTrue(self.task.supports_deb822_sources("xenial"))
+
+    def test_supports_inline_signed_by(self) -> None:
+        """Test supports_inline_signed_by with 4 sample distributions."""
+        self.assertFalse(self.task.supports_deb822_sources("jessie"))
+        self.assertFalse(self.task.supports_deb822_sources("trusty"))
+        self.assertTrue(self.task.supports_deb822_sources("bookworm"))
+        self.assertTrue(self.task.supports_deb822_sources("jammy"))
+
+    def test_write_extra_repository_config_jessie_no_keys(self) -> None:
+        """Test repositories without deb822 or keys."""
+        self.configure_extra_repositories(
+            {
+                "url": "http://example.net",
+                "suite": "bookworm",
+                "components": ["main"],
+            },
+        )
+
+        self.task.write_extra_repository_config("jessie", self.directory)
+
+        self.assertEqual(self.task.extra_repository_keys, [])
+        self.assertEqual(self.task.extra_repository_sources, [])
+
+    def test_write_extra_repository_config_two_keys(self) -> None:
+        """Test repositories with two keys."""
+        self.configure_extra_repositories(
+            {
+                "url": "http://example.net/",
+                "suite": "a",
+                "components": ["a"],
+                "signing_key": "KEY A",
+            },
+            {
+                "url": "http://example.net/",
+                "suite": "b/",
+                "signing_key": "KEY B",
+            },
+        )
+
+        self.task.write_extra_repository_config("jessie", self.directory)
+
+        self.assertEqual(len(self.task.extra_repository_keys), 2)
+        expected_keys = ["KEY A\n", "KEY B\n"]
+        for key_file, expected_key in zip(
+            self.task.extra_repository_keys, expected_keys
+        ):
+            self.assertEqual(key_file.read_text(), expected_key)
+
+        self.assertEqual(self.task.extra_repository_sources, [])
+
+    def test_write_extra_repository_config_deb822_signed_keyring(self) -> None:
+        """Test repositories in deb822 format with keys in apt < 2.3.10."""
+        self.configure_extra_repositories(
+            {
+                "url": "http://example.net",
+                "suite": "bullseye",
+                "components": ["main", "non-free"],
+                "signing_key": "\n".join(
+                    (
+                        "-----BEGIN PGP PUBLIC KEY BLOCK-----",
+                        "",
+                        "ABCDEFGHI",
+                        "-----END PGP PUBLIC KEY BLOCK-----",
+                    )
+                ),
+            },
+        )
+
+        self.task.write_extra_repository_config("bullseye", self.directory)
+
+        self.assertEqual(len(self.task.extra_repository_keys), 1)
+        self.assertEqual(
+            self.task.extra_repository_keys[0].read_text(),
+            textwrap.dedent(
+                """\
+                -----BEGIN PGP PUBLIC KEY BLOCK-----
+
+                ABCDEFGHI
+                -----END PGP PUBLIC KEY BLOCK-----
+                """
+            ),
+        )
+        self.assertEqual(
+            self.task.extra_repository_keys[0].name,
+            "extra_apt_key_0.asc",
+        )
+        self.assertEqual(len(self.task.extra_repository_sources), 1)
+        self.assertEqual(
+            self.task.extra_repository_sources[0].read_text(),
+            textwrap.dedent(
+                """\
+                Types: deb
+                URIs: http://example.net
+                Suites: bullseye
+                Components: main non-free
+                Signed-By: /etc/apt/keyrings/extra_apt_key_0.asc
+                """
+            ),
+        )
+
+    def test_write_extra_repository_config_deb822_signed_inline(self) -> None:
+        """Test repositories in deb822 format with keys inline."""
+        self.configure_extra_repositories(
+            {
+                "url": "http://example.net",
+                "suite": "bookworm",
+                "components": ["main", "non-free"],
+                "signing_key": "\n".join(
+                    (
+                        "-----BEGIN PGP PUBLIC KEY BLOCK-----",
+                        "",
+                        "ABCDEFGHI",
+                        "-----END PGP PUBLIC KEY BLOCK-----",
+                    )
+                ),
+            },
+        )
+
+        self.task.write_extra_repository_config("bookworm", self.directory)
+
+        self.assertEqual(self.task.extra_repository_keys, [])
+        self.assertEqual(len(self.task.extra_repository_sources), 1)
+        self.assertEqual(
+            self.task.extra_repository_sources[0].read_text(),
+            textwrap.dedent(
+                """\
+                Types: deb
+                URIs: http://example.net
+                Suites: bookworm
+                Components: main non-free
+                Signed-By:
+                 -----BEGIN PGP PUBLIC KEY BLOCK-----
+                 .
+                 ABCDEFGHI
+                 -----END PGP PUBLIC KEY BLOCK-----
+                """
+            ),
+        )
+
+    def test_iter_oneline_sources(self) -> None:
+        """Test iter_oneline_sources with a variety of sources."""
+        self.configure_extra_repositories(
+            {
+                "url": "http://example.net",
+                "suite": "flat/",
+            },
+            {
+                "url": "http://example.com",
+                "suite": "bookworm",
+                "components": ["main", "non-free"],
+            },
+            {
+                "url": "http://example.org",
+                "suite": "buster",
+                "components": ["main", "non-free"],
+                "signing_key": "\n".join(
+                    (
+                        "-----BEGIN PGP PUBLIC KEY BLOCK-----",
+                        "",
+                        "ABCDEFGHI",
+                        "-----END PGP PUBLIC KEY BLOCK-----",
+                    )
+                ),
+            },
+        )
+        sources = list(self.task.iter_oneline_sources())
+        self.assertEqual(
+            sources,
+            [
+                "deb http://example.net flat/",
+                "deb http://example.com bookworm main non-free",
+                (
+                    "deb [signed-by=/etc/apt/keyrings/extra_apt_key_2.asc] "
+                    "http://example.org buster main non-free"
+                ),
+            ],
+        )
